@@ -27,7 +27,43 @@ Features:
 """
 from collections import defaultdict
 import math
-import numpy as np
+import numpy
+
+
+# ---------------- Enable switching between NumPy and CuPy dynamically.
+
+
+array_library = numpy  # numpy or cupy
+
+
+class ArrayLibraryProxy:
+    """Enable switching between NumPy and CuPy dynamically.
+    
+    E.g.
+    import smallpebble as sp
+    import cupy
+
+    sp.array_library = cupy
+    """
+
+    def __getattribute__(self, name):
+        return getattr(array_library, name)
+
+
+def np_add_at(a, indices, b):
+    """Apply either np.add.at or cupy.scatter_add (which are equivalent),
+    depending on which library is being used. 
+    Do this because CuPy has no cupy.add.at.
+    """
+    if array_library.__name__ == "numpy":
+        return array_library.add.at(a, indices, b)
+    elif array_library.__name__ == "cupy":
+        return array_library.scatter_add(a, indices, b)
+    else:
+        raise ValueError("Expected array_library.__name__ to be `numpy` or `cupy`.")
+
+
+np = ArrayLibraryProxy()
 
 
 # ----------------
@@ -80,9 +116,10 @@ def add(a, b):
 
 def add_at(a, indices, b):
     """Add the elements of `b` to the locations in `a` specified by `indices`.
-    Allows adding to an element of `a` repeatedly."""
+    Allows adding to an element of `a` repeatedly.
+    """
     value = a.array.copy()
-    np.add.at(value, indices, b.array)
+    np_add_at(value, indices, b.array)
     local_gradients = (
         (a, lambda path_value: path_value),
         (b, lambda path_value: getitem(path_value, indices)),
@@ -340,25 +377,13 @@ def conv2d(images, kernels, padding="SAME", strides=[1, 1]):
     n_images, imheight, imwidth, _ = images.array.shape
     kernheight, kernwidth, channels_in, channels_out = kernels.array.shape
     stride_y, stride_x = strides
-
-    if padding == "SAME":
-        pad_top, pad_bottom = pad_amounts(imheight, stride_y, kernheight)
-        pad_left, pad_right = pad_amounts(imwidth, stride_x, kernwidth)
-        images = pad(
-            images,
-            pad_width=[(0, 0), (pad_top, pad_bottom), (pad_left, pad_right), (0, 0),],
-        )
-        _, imheight, imwidth, _ = images.array.shape
-    elif padding == "VALID":
-        pass
-    else:
-        raise ValueError("padding must be 'SAME' or 'VALID'.")
-
+    images, imheight, imwidth = padding2d(
+        images, padding, imheight, imwidth, stride_y, stride_x, kernheight, kernwidth
+    )
     # Index to get the image patches:
     index_of_patches, outheight, outwidth, n_patches = patches_index(
         imheight, imwidth, kernheight, kernwidth, stride_y, stride_x
     )
-
     #  Extract patches, and reshape so can matrix multiply.
     patches = getitem(
         images, (slice(None), index_of_patches[0], index_of_patches[1], slice(None)),
@@ -369,7 +394,6 @@ def conv2d(images, kernels, padding="SAME", strides=[1, 1]):
     kernels_as_matrix = reshape(
         kernels, [kernheight * kernwidth * channels_in, channels_out]
     )
-
     result = matmul(patches_as_matrix, kernels_as_matrix)
     return reshape(result, [n_images, outheight, outwidth, channels_out])
 
@@ -382,7 +406,23 @@ def maxpool2d(images, kernheight, kernwidth, padding="SAME", strides=[1, 1]):
     """Maxpooling on a `Variable` of shape [n_images, imheight, imwidth, n_channels]."""
     n_images, imheight, imwidth, n_channels = images.array.shape
     stride_y, stride_x = strides
+    images, imheight, imwidth = padding2d(
+        images, padding, imheight, imwidth, stride_y, stride_x, kernheight, kernwidth
+    )
+    index_of_patches, outheight, outwidth, n_patches = patches_index(
+        imheight, imwidth, kernheight, kernwidth, stride_y, stride_x
+    )
+    patches = getitem(
+        images, (slice(None), index_of_patches[0], index_of_patches[1], slice(None)),
+    )
+    patches_max = maxax(patches, axis=2)
+    return reshape(patches_max, [n_images, outheight, outwidth, n_channels])
 
+
+def padding2d(
+    images, padding, imheight, imwidth, stride_y, stride_x, kernheight, kernwidth
+):
+    """Pad `images` for conv2d, maxpool2d."""
     if padding == "SAME":
         pad_top, pad_bottom = pad_amounts(imheight, stride_y, kernheight)
         pad_left, pad_right = pad_amounts(imwidth, stride_x, kernwidth)
@@ -395,15 +435,7 @@ def maxpool2d(images, kernheight, kernwidth, padding="SAME", strides=[1, 1]):
         pass
     else:
         raise ValueError("padding must be 'SAME' or 'VALID'.")
-
-    index_of_patches, outheight, outwidth, n_patches = patches_index(
-        imheight, imwidth, kernheight, kernwidth, stride_y, stride_x
-    )
-    patches = getitem(
-        images, (slice(None), index_of_patches[0], index_of_patches[1], slice(None)),
-    )
-    patches_max = maxax(patches, axis=2)
-    return reshape(patches_max, [n_images, outheight, outwidth, n_channels])
+    return images, imheight, imwidth
 
 
 # ---------------- UTIL
@@ -469,11 +501,58 @@ Variable.__mul__ = mul
 Variable.__sub__ = sub
 Variable.__truediv__ = div
 
-# Indexing:
+# Indexing (no __setitem__, use `setat` instead):
 Variable.__getitem__ = getitem
-# No __setitem__, use `setat`.
 
 # Useful attributes:
 Variable.dtype = property(lambda self: self.array.dtype)
 Variable.ndim = property(lambda self: self.array.ndim)
 Variable.shape = property(lambda self: self.array.shape)
+
+
+# ----------------
+# ---------------- DEPRECATED
+# ----------------
+# May or may not be removed at some point. Moved here to reduce noise.
+
+
+def bincount(idx, a, size):
+    """Faster than `np.add.at`. `idx`: flat NumPy array , `a` flat Variable, `size` int.
+    
+    Deprecation reason:
+    add.at is the function we want really.
+    Although bincount is faster when using numpy, it probably isn't worth the obfuscation.
+    Also, CuPy doesn't benefit from the bincount version of add.at,
+    since cupy.scatter_add is faster.
+    """
+    value = np.bincount(idx, a.array, minlength=size)
+    local_gradients = ((a, lambda path_value: getitemflat(path_value, idx)),)
+    return Variable(value, local_gradients)
+
+
+def getitemflat(a, idx):
+    """`a` is flat, `idx` is row major (and flat).
+    
+    Deprecation reason:
+    This is the sister function of `sp.bincount`, and is deprecated for the same reason.
+    """
+    value = a.array[idx]
+    local_gradients = ((a, lambda path_value: bincount(idx, path_value, a.array.size)),)
+    return Variable(value, local_gradients)
+
+
+def sliding_window_view(a, window_shape):
+    value = np.lib.stride_tricks.sliding_window_view(a.array, window_shape)
+
+    def multiply_by_locgrad(path_value):
+        # this is generally quicker than add.at ..
+        idx = np.arange(a.array.size).reshape(a.array.shape)
+        idx = np.lib.stride_tricks.sliding_window_view(idx, window_shape)
+        idx = idx.reshape(-1)
+
+        path_value = reshape(path_value, (-1))
+        return reshape(bincount(idx, path_value, a.array.size), a.array.shape)
+
+    local_gradients = ((a, multiply_by_locgrad),)
+
+    return Variable(value, local_gradients)
